import logging
import hashlib
import json

import requests

import bpy

from render_polargrid.utils import get_settings


logger = logging.getLogger(__name__)


def calculate_file_hash(path):
    logger.debug("calculate_file_hash {{ path: {} }}".format(path))
    sha = hashlib.sha512()

    with open(str(path), "rb") as handle:
        buf = handle.read(4096)
        while buf != b'':
            buf = handle.read(4096)
            sha.update(buf)
    return sha.hexdigest()


class Client():
    def __init__(self, context, reporter):
        self.__context = context
        self.__settings = get_settings(context)
        self.__reporter = reporter

        # Additional kwargs send to each request.
        self.__kwargs = {}

    def _check_response(self, response):
        if response.status_code != 200:
            message = "Did not receive a correct response from the polargrid server (received status code: {})".format(response.status_code)
            logger.error(message)
            self.__reporter.report({'ERROR'}, message)

    def request_token(self, email, token, secret):
        url = self.__settings.backend_url + "api/request_token/"
        response = requests.post(url, json={
            'email': email,
            'token': token,
            'secret': secret,
        }, **self.__kwargs)
        self._check_response(response)

    def validate_token(self, email, token):
        url = self.__settings.backend_url + "api/validate_token/"
        response = requests.post(url, json={
            'email': email,
            'token': token,
        }, **self.__kwargs)
        self._check_response(response)
        return response.json()

    def list_server_pools(self):
        url = self.__settings.backend_url + "api/user/server_pools/"
        response = requests.get(url, json={
            'token': self.__settings.token,
            'secret': self.__settings.secret,
        }, **self.__kwargs)
        self._check_response(response)
        return response.json()

    def modify_scene(self, scene, status):
        assert(status in ['Locked', 'Ok'])
        project_root = scene.polargrid.get_project_root(self.__context)
        scene_root = scene.polargrid.get_scene_root(self.__context)
        base_name = bpy.path.basename(self.__context.blend_data.filepath)
        abs_path = scene_root / base_name
        rel_name = str(abs_path.relative_to(project_root))

        url = self.__settings.backend_url + "api/scenes/"
        response = requests.post(url, json={
            'email': self.__settings.email,
            'token': self.__settings.token,
            'secret': self.__settings.secret,
            'scene_id': scene.polargrid.scene_id,
            'name': rel_name,
            'scene_status': status,
        }, **self.__kwargs)
        self._check_response(response)
        return response.json()

    def open_polargrid(self):
        # TODO: use api service to receive tmp and use this in the browser.
        # this key will have a timespan of one hour. for requests.
        url = "{base_url}?token={token}&secret={secret}".format(
            base_url=self.__settings.backend_url.replace("8080", "8000"),
            token=self.__settings.token,
            secret=self.__settings.secret,
        )
        import webbrowser
        webbrowser.open(url)

    def create_job(self, scene, job_type, priority, start_frame, end_frame, use_flip_fluids: bool):
        assert(job_type in ['RenderStill', 'RenderAnimation', 'BakeFlipFluids'])
        # pool_code = self.__context.window_manager.polargrid.pool;
        pool = self.__context.window_manager.polargrid._get_pool()
        use_gpu = pool['is_gpu_pool']
        url = self.__settings.backend_url + "api/jobs/"

        # Build a list of required addons
        addons = []
        if use_flip_fluids:
            addons.append('FlipFluids')

        response = requests.post(url, json={
            'email': self.__settings.email,
            'token': self.__settings.token,
            'secret': self.__settings.secret,
            'scene_id': scene.polargrid.scene_id,

            'application': {
                'name': "Blender",
                'version': get_app_version(),
            },
            'options': {
                'use_device_gpu': use_gpu,
                'use_device_cpu': not use_gpu,
                'output_file_type': scene.render.image_settings.file_format,
                'required_addons': addons,
            },
            'job_type': job_type,
            'pool_code': pool['code'],

            'priority': priority,
            'start_frame': start_frame,
            'end_frame': end_frame if job_type in ['RenderAnimation', 'BakeFlipFluids'] else None,
        }, **self.__kwargs)
        self._check_response(response)
        return response.json()

    def _send_file(self, src, dst, file_hash, callback) -> bool:
        logger.info(f"uploading_file: {dst}")
        url = self.__settings.backend_url + "api/files/"
        body = {
            'auth': {
                'token': self.__settings.token,
                'secret': self.__settings.secret,
            },
            'scene_id': self.__context.scene.polargrid.scene_id,
            'path': str(dst).replace("\\", "/"),
            'file_type': 'Input',
            'hash': file_hash,
        }
        response = requests.post(url, data=ReportReader(filename=str(src), callback=callback), headers={
            'POLARGRID-REQUEST-DATA': json.dumps(body)
        }, **self.__kwargs)
        self._check_response(response)
        return True

    def _send_files(self, files_to_upload: list, file_hashes: dict, callback) -> bool:
        for src, dst in files_to_upload:
            file_hash = file_hashes[src]
            if not self._send_file(src, dst, file_hash, callback):
                return False
        return True

    def _retrieve_file_hashes(self, files_to_upload: list) -> dict:
        """
        Retrieve the file hashes on the server.
        """
        url = self.__settings.backend_url + "api/files/hashes/"
        request_data = {
            'auth': {
                'token': self.__settings.token,
                'secret': self.__settings.secret,
            },
            'scene_id': self.__context.scene.polargrid.scene_id,
            'file_type': 'Input',
            'files': [{
                'file_path': str(file_to_upload[1]).replace("\\", "/")
            } for file_to_upload in files_to_upload]
        }
        response = requests.get(url, json=request_data, **self.__kwargs)
        self._check_response(response)
        result = {}
        for (src, dst), (file_pair) in zip(files_to_upload, response.json()):
            result[src] = file_pair['file_hash']
        return result

    def init_upload_files(self) -> dict:
        """
        Initialize uploading of multiple files.
        The result of this function has all know server hashes.
        it needs to be passed the `upload_files` method.

        The reason for this split it to provide better feedback to the
        user which file is currently checked and uploaded. It also
        removes the stall at the beginning of the upload process
        as the local file hashes don't need to be calculated upfront.
        """
        logger.debug("retrieve_file_hashes_from_server")
        url = self.__settings.backend_url + "api/files/hashes/"
        request_data = {
            'auth': {
                'token': self.__settings.token,
                'secret': self.__settings.secret,
            },
            'scene_id': self.__context.scene.polargrid.scene_id,
            'file_type': 'Input',
            'files': []
        }
        response = requests.get(url, json=request_data, **self.__kwargs)
        self._check_response(response)
        result = {
            file_pair['file_path']: file_pair['file_hash'] for file_pair in response.json()
        }
        return result

    def upload_file(self, file_to_upload: tuple, server_file_hashes: dict, callback, force:bool=False) -> bool:
        """
        Upload a single file, with prefetched server hashes.

        The server_file_hashes is the result of `init_upload_files`

        @param file_to_upload: tuple<str,str> where the tuple is a src
            dst pair.
        """
        (src, dst) = file_to_upload

        file_hash = calculate_file_hash(src)
        server_file_hash = None if force else server_file_hashes.get(str(dst))
        if file_hash == server_file_hash:
            logger.debug(f"same hash on server skipping upload {{ file: {dst} }}")
            callback(src.stat().st_size)
            return True
        else:
            return self._send_file(src, dst, file_hash, callback)

    def upload_files(self, files_to_upload: list, callback, force:bool=False) -> bool:
        """
        Upload multiple files.

        This method isn't in use by the blender addon as the progress reporting
        looses some detail due to how BAT is setup.

        @param files_to_upload: list<tuple<str,str>> where the tuple is a src
            dst pair.
        """
        server_file_hashes = self._retrieve_file_hashes(files_to_upload)

        if force:
            calculated_file_hashes = {
                file_pair[0]: calculate_file_hash(file_pair[0]) for file_pair in files_to_upload
            }
            return self._send_files(files_to_upload, calculated_file_hashes, callback)
        changed_files = []
        for src, dst in files_to_upload:
            file_hash = calculate_file_hash(src)
            server_file_hash = server_file_hashes[src]
            if file_hash == server_file_hash:
                logger.debug(f"same hash on server skipping upload {{ file: {dst} }}")
                callback(src.stat().st_size)
                continue
            changed_files.append((src, dst))
        return self._send_files(changed_files, calculated_file_hashes, callback)


class ReportReader(object):
    def __init__(self, filename, callback, chunksize=1 << 13):
        self.filename = filename
        self.chunksize = chunksize
        self.callback = callback

    def __iter__(self):
        with open(self.filename, 'rb') as file:
            while True:
                data = file.read(self.chunksize)
                if not data:
                    break
                self.callback(len(data))
                yield data


def get_app_version():
    version = bpy.app.version
    return version[0]*100 + version[1]
