import os
from pathlib import Path
import logging

import bpy

from render_polargrid.polargrid_client import Client
from render_polargrid.polargrid_upload import copy, abort, Aborted, FileTransferError
from render_polargrid.validation import is_preferences_valid, is_valid_project_root
import render_polargrid.version_compatibility_utils as vcu
from render_polargrid import async_loop
from blender_asset_tracer.bpathlib import BlendPath


logger = logging.getLogger(__name__)

class PGRenderBasePanelMixin(object):
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_label = "PolarGrid Render"

    def draw(self, context):
        layout = self.layout
        wm = context.window_manager
        pg = wm.polargrid

        if pg.status not in ['IDLE', 'DONE', 'ABORTED']:
            row = layout.row()
            row.enabled=False
            row.prop(pg, 'progress', text=pg.description)

        else:
            if pg._get_pool():
                split = vcu.ui_split(ui_element=layout, factor=0.9)
                split.prop(pg, 'pool')
                split.operator('polargrid.fetch_pools', text="", icon='FILE_REFRESH')

                self._draw_operations(context, layout)
            else:
                layout.operator('polargrid.fetch_pools', icon='FILE_REFRESH')

        layout.operator('polargrid.view_jobs', icon='URL')

    def _draw_operations(self, context, layout):
        """
        Method where subclasses can place their render operations
        """
        raise NotImplementedError()



class PGRenderPanel(bpy.types.Panel, PGRenderBasePanelMixin):
    bl_idname = "POLARGRID_PT_render"
    bl_context = "render"

    COMPAT_ENGINES = {'CYCLES'}

    @classmethod
    def poll(cls, context):
        rd = context.scene.render
        return rd.engine in cls.COMPAT_ENGINES

    def _draw_operations(self, context, layout):
        op = layout.operator('polargrid.render', text="Render Frame", icon='RENDER_STILL')
        op.job_type = 'RenderStill'
        op.priority = 50
        op.start_frame = context.scene.frame_current

        op = layout.operator('polargrid.render', text="Render Animation", icon='RENDER_ANIMATION')
        op.job_type = 'RenderAnimation'
        op.priority = 50
        op.start_frame = context.scene.frame_start
        op.end_frame = context.scene.frame_end


class PGPhysicsPanel(bpy.types.Panel, PGRenderBasePanelMixin):
    bl_idname = "POLARGRID_PT_physics"
    bl_context = "physics"
    bl_category = "FLIP Fluid"

    @classmethod
    def poll(self, context):
        obj = context.active_object

        if not hasattr(obj, 'flip_fluid'):
            return False
        result = obj.flip_fluid.is_domain()
        return result

    def _draw_operations(self, context, layout):
        pool = context.window_manager.polargrid._get_pool()
        row = layout.row()
        row.enabled = "FlipFluids" in pool['features']
        op = row.operator('polargrid.render', text="Bake On Polargrid")
        op.job_type = 'BakeFlipFluids'
        op.priority = 50
        op.start_frame = context.scene.frame_start
        op.end_frame = context.scene.frame_end


PG_JOB_TYPES = (
    ("RenderStill",     "Render Current Frame", ""),
    ("RenderAnimation", "Render Animation",     ""),
    ("BakeFlipFluids",  "Bake Flip Fluids",     ""),
)

class BackupSceneSettings:
    def __init__(self, scene:bpy.types.Scene):
        self.scene = scene
    def __enter__(self):
        self._backup()
    def __exit__(self, *args, **kwargs):
        self._restore()

    def _backup(self):
        self.cycles_device = self.scene.cycles.device

        render = self.scene.render
        self.use_file_extension = render.use_file_extension
        self.use_overwrite = render.use_overwrite
        self.use_placeholder = render.use_placeholder
        if hasattr(render, "tile_x"):
            self.tile_x = render.tile_x
            self.tile_y = render.tile_y

    def _restore(self):
        self.scene.cycles.device = self.cycles_device

        render = self.scene.render
        render.use_file_extension = self.use_file_extension
        render.use_overwrite = self.use_overwrite
        render.use_placeholder = self.use_placeholder
        if hasattr(render, "tile_x"):
            render.tile_x = self.tile_x
            render.tile_y = self.tile_y


class PGRenderOperator(async_loop.AsyncModalOperatorMixin, bpy.types.Operator):
    bl_idname = "polargrid.render"
    bl_label = "Render on PolarGrid"
    stop_upon_exception = True

    job_type = bpy.props.EnumProperty(
        items=PG_JOB_TYPES,
        name="Job Type",
        default="RenderStill",
    )
    exec(vcu.convert_attribute_to_28("job_type"))
    priority = bpy.props.IntProperty(
        min=0,
        max=32000,
        default=50,
    )
    exec(vcu.convert_attribute_to_28("priority"))
    start_frame = bpy.props.IntProperty(
        default=1,
    )
    exec(vcu.convert_attribute_to_28("start_frame"))
    end_frame = bpy.props.IntProperty(
        default=250,
    )
    exec(vcu.convert_attribute_to_28("end_frame"))

    async def async_execute(self, context):
        if is_preferences_valid(context):
            context.scene.polargrid._ensure_scene_id()

            client = Client(context=context, reporter=self)

            current_filename = context.blend_data.filepath

            if not os.path.exists(current_filename):
                self.report({'ERROR'}, 'Please save your Blend file before using '
                                       'the Polargrid addon.')
                self.quit()
                return


            if current_filename == "":
                reporter.report({'ERROR'}, "Current scene is not saved. Please save your scene and try again!")
                # no file loaded
                return {'CANCELED'}

            if self.job_type == 'BakeFlipFluids':
                obj = context.active_object
                if not obj.flip_fluid.is_domain():
                    self.report({'ERROR'}, "Active object is not a domain object")
                    return {'CANCELED'}

            # create a tmp file (currentfilename + .polargrid)
            context.window_manager.polargrid.status = 'SAVING'
            bfile = await self._save_blendfile(context)
            target_dirname = current_filename + ".polargrid"

            project_path = context.scene.polargrid.get_project_root(context)
            target_path = Path("/")

            # For now we assume that all libraries needs to be packed in order to work around
            # the packed blend file limitation in BAT. Best option would be to support packed
            # blend files in BAT.
            has_packed_libraries = (len(bpy.data.libraries) and all([lib.packed_file for lib in bpy.data.libraries]))
            try:
                outfile, missing_sources, use_flip_fluids = await copy(
                    context=bpy.context,
                    base_blendfile=bfile,
                    project=project_path,
                    target=target_path,
                    exclusion_filter="",
                    has_packed_libraries=has_packed_libraries,
                    client=client)

                client.create_job(
                    scene=context.scene,
                    job_type=self.job_type,
                    priority=self.priority,
                    start_frame=self.start_frame,
                    end_frame=self.end_frame,
                    use_flip_fluids=use_flip_fluids,
                )
            except FileTransferError as ex:
                logger.error('Could not transfer %d files, starting with %s',
                               len(ex.files_remaining), ex.files_remaining[0])
                self.report({'ERROR'}, 'Unable to transfer %d files' % len(ex.files_remaining))
                self.quit()
                return {'CANCELLED'}
            except Aborted:
                logger.warning('BAT Pack was aborted')
                self.report({'WARNING'}, 'Aborted Polargrid file packing/transferring')
                self.quit()
                return {'CANCELLED'}

            bfile.unlink()
        else:
            return bpy.ops.wm.addon_userpref_show(module='render_polargrid')

        return {'FINISHED'}

    def quit(self):
        if bpy.context.window_manager.polargrid.status != 'ABORTED':
            bpy.context.window_manager.polargrid.status = 'DONE'
        super().quit()

    async def _save_blendfile(self, context):
        """Save to a different file, specifically for Polargrid.

        We shouldn't overwrite the artist's file.
        We can compress, since this file won't be managed by SVN and doesn't need diffability.
        """
        scene = context.scene
        render = scene.render
        obj = context.active_object

        # Remember settings we need to restore after saving.
        with BackupSceneSettings(scene):

            if self.job_type == 'BakeFlipFluids':
                if obj.flip_fluid.is_domain():
                    old_flip_fluid_domain_cache_cache_directory = obj.flip_fluid.domain.cache.cache_directory
                    obj.flip_fluid.domain.cache.cache_directory = "//flip_fluids_cache/"

            try:
                pool = context.window_manager.polargrid._get_pool()
                use_gpu = pool['is_gpu_pool']
                scene.cycles.device = 'GPU' if use_gpu else 'CPU'
                # Pre blender 3.0 setting.
                if hasattr(render, "tile_x"):
                    render.tile_x = 256 if use_gpu else 32
                    render.tile_y = 256 if use_gpu else 32
                # The file extension should be determined by the render settings, not necessarily
                # by the setttings in the output panel.
                render.use_file_extension = True

                # Rescheduling should not overwrite existing frames.
                render.use_overwrite = False
                render.use_placeholder = False

                filepath = Path(context.blend_data.filepath).with_suffix('.polar')
                logger.info('Saving copy to temporary file %s', filepath)
                bpy.ops.wm.save_as_mainfile(filepath=str(filepath),
                                            compress=True,
                                            copy=True)
            finally:
                # Restore the settings we changed, even after an exception.
                if self.job_type == 'BakeFlipFluids':
                    if obj.flip_fluid.is_domain():
                        obj.flip_fluid.domain.cache.cache_directory = old_flip_fluid_domain_cache_cache_directory

        return filepath

class PGViewJobsOperator(bpy.types.Operator):
    bl_idname = "polargrid.view_jobs"
    bl_label = "View Jobs"

    def execute(self, context):
        if is_preferences_valid(context):
            client = Client(context=context, reporter=self)
            client.open_polargrid()

        else:
            return bpy.ops.wm.addon_userpref_show(module='render_polargrid')

        return {'FINISHED'}


class PGScenePanel(bpy.types.Panel):
    bl_idname = "POLARGRID_PT_scene"
    bl_label = "PolarGrid Render"
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "scene"

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        pg = scene.polargrid
        row = layout.row()
        row.enabled = False
        row.prop(pg, 'scene_id')
        layout.operator('polargrid.new_scene_id')

        icon_id = 'NONE'
        if not is_valid_project_root(pg.project_root):
            icon_id='ERROR'
        layout.prop(pg, 'project_root', icon=icon_id)


class PGNewSceneIDOperator(bpy.types.Operator):
    bl_idname = "polargrid.new_scene_id"
    bl_label = "Generate New Scene ID"

    @classmethod
    def poll(cls, context):
        return context.scene is not None

    def execute(self, context):
        context.scene.polargrid._new_scene_id()
        return {'FINISHED'}
